@@ -1,6 +1,3 @@
-;;;; $Id$
-;;;; $URL$
-
 ;;;; New ABCL networking support (replacement to old armedbear.lisp)
 ;;;; Author: Chun Tian (binghe)
 
@@ -62,6 +59,8 @@
 (defvar $@close/Selector/0 (jmethod $*Selector "close"))
 (defvar $@close/ServerSocket/0 (jmethod $*ServerSocket "close"))
 (defvar $@close/Socket/0 (jmethod $*Socket "close"))
+(defvar $@shutdownInput/Socket/0 (jmethod $*Socket "shutdownInput"))
+(defvar $@shutdownOutput/Socket/0 (jmethod $*Socket "shutdownOutput"))
 (defvar $@configureBlocking/1 (jmethod $*SelectableChannel "configureBlocking" $*boolean))
 (defvar $@connect/DatagramChannel/1 (jmethod $*DatagramChannel "connect" $*SocketAddress))
 (defvar $@connect/Socket/1 (jmethod $*Socket "connect" $*SocketAddress))
@@ -173,7 +172,9 @@
            (vector (jbyte 0) (jbyte 1) (jbyte 2) (jbyte 3)))
           ((= 16 length)
            (vector (jbyte 0) (jbyte 1) (jbyte 2) (jbyte 3) 
-                   (jbyte 4) (jbyte 5) (jbyte 6) (jbyte 7)))
+                   (jbyte 4) (jbyte 5) (jbyte 6) (jbyte 7)
+                   (jbyte 8) (jbyte 9) (jbyte 10) (jbyte 11)
+                   (jbyte 12) (jbyte 13) (jbyte 14) (jbyte 15)))
           (t nil)))))) ; neither a IPv4 nor IPv6 address?!
 
 (defun get-hosts-by-name (name)
@@ -264,10 +265,6 @@
 
 ;;; SOCKET-CLOSE
 
-(defmethod socket-close :before ((usocket usocket))
-  (when (wait-list usocket)
-     (remove-waiter (wait-list usocket) usocket)))
-
 (defmethod socket-close ((usocket stream-server-usocket))
   (with-mapped-conditions (usocket)
     (jcall $@close/ServerSocket/0 (socket usocket))))
@@ -281,6 +278,14 @@
   (with-mapped-conditions (usocket)
     (jcall $@close/DatagramSocket/0 (socket usocket))))
 
+(defmethod socket-shutdown ((usocket stream-usocket) direction)
+  (with-mapped-conditions (usocket)
+    (ecase direction
+      (:input
+       (jcall $@shutdownInput/Socket/0 (socket usocket)))
+      (:output
+       (jcall $@shutdownOutput/Socket/0 (socket usocket))))))
+
 ;;; GET-LOCAL/PEER-NAME/ADDRESS/PORT
 
 (defmethod get-local-name ((usocket usocket))
