@@ -104,11 +104,13 @@
 ;; because socket-streams are also sockets.
 (defmethod socket-close ((usocket usocket))
   "Close socket."
-  (when (wait-list usocket)
-     (remove-waiter (wait-list usocket) usocket))
   (with-mapped-conditions (usocket)
     (close (socket usocket))))
 
+(defmethod socket-shutdown ((usocket stream-usocket) direction)
+  (with-mapped-conditions (usocket)
+    (socket:shutdown (socket usocket) :direction direction)))
+
 (defun socket-listen (host port
                            &key reuseaddress
                            (reuse-address nil reuse-address-supplied-p)
