@@ -1,6 +1,3 @@
-;;;; $Id$
-;;;; $URL$
-
 ;;;; See LICENSE for licensing information.
 
 (in-package :usocket)
@@ -9,7 +6,7 @@
   (require "comm")
 
   #+lispworks3
-  (error "LispWorks 3 is not supported by USOCKET any more."))
+  (error "LispWorks 3 is not supported"))
 
 ;;; ---------------------------------------------------------------------------
 ;;;  Warn if multiprocessing is not running on Lispworks
@@ -65,7 +62,7 @@
     (if usock-err
         (if (subtypep usock-err 'error)
             (error usock-err :socket socket)
-          (signal usock-err :socket socket))
+          (signal usock-err))
       (error 'unknown-error
              :socket socket
              :real-error condition
@@ -89,6 +86,11 @@
   #+linux 20
   "Socket receive timeout")
 
+(defconstant *sockopt_so_sndtimeo*
+  #-linux #x1007
+  #+linux 21
+  "Socket send timeout")
+
 (fli:define-c-struct timeval
   (tv-sec :long)
   (tv-usec :long))
@@ -139,6 +141,24 @@
                                (fli:size-of '(:struct timeval))))
             seconds)))))
 
+#-win32
+(defun set-socket-send-timeout (socket-fd seconds)
+  "Set socket option: SNDTIMEO, argument seconds can be a float number"
+  (declare (type integer socket-fd)
+           (type number seconds))
+  (multiple-value-bind (sec usec) (truncate seconds)
+    (fli:with-dynamic-foreign-objects ((timeout (:struct timeval)))
+      (fli:with-foreign-slots (tv-sec tv-usec) timeout
+        (setf tv-sec sec
+              tv-usec (truncate (* 1000000 usec)))
+        (if (zerop (comm::setsockopt socket-fd
+                               comm::*sockopt_sol_socket*
+                               *sockopt_so_sndtimeo*
+                               (fli:copy-pointer timeout
+                                                 :type '(:pointer :void))
+                               (fli:size-of '(:struct timeval))))
+            seconds)))))
+
 #+win32
 (defun set-socket-receive-timeout (socket-fd seconds)
   "Set socket option: RCVTIMEO, argument seconds can be a float number.
@@ -156,6 +176,23 @@
                            (fli:size-of :int)))
         seconds)))
 
+#+win32
+(defun set-socket-send-timeout (socket-fd seconds)
+  "Set socket option: SNDTIMEO, argument seconds can be a float number.
+   On win32, you must bind the socket before use this function."
+  (declare (type integer socket-fd)
+           (type number seconds))
+  (fli:with-dynamic-foreign-objects ((timeout :int))
+    (setf (fli:dereference timeout)
+          (truncate (* 1000 seconds)))
+    (if (zerop (comm::setsockopt socket-fd
+                           comm::*sockopt_sol_socket*
+                           *sockopt_so_sndtimeo*
+                           (fli:copy-pointer timeout
+                                             :type '(:pointer :char))
+                           (fli:size-of :int)))
+        seconds)))
+
 #-win32
 (defun get-socket-receive-timeout (socket-fd)
   "Get socket option: RCVTIMEO, return value is a float number"
@@ -171,6 +208,21 @@
     (fli:with-foreign-slots (tv-sec tv-usec) timeout
       (float (+ tv-sec (/ tv-usec 1000000))))))
 
+#-win32
+(defun get-socket-send-timeout (socket-fd)
+  "Get socket option: SNDTIMEO, return value is a float number"
+  (declare (type integer socket-fd))
+  (fli:with-dynamic-foreign-objects ((timeout (:struct timeval))
+                                     (len :int))
+    (comm::getsockopt socket-fd
+                comm::*sockopt_sol_socket*
+                *sockopt_so_sndtimeo*
+                (fli:copy-pointer timeout
+                                  :type '(:pointer :void))
+                len)
+    (fli:with-foreign-slots (tv-sec tv-usec) timeout
+      (float (+ tv-sec (/ tv-usec 1000000))))))
+
 #+win32
 (defun get-socket-receive-timeout (socket-fd)
   "Get socket option: RCVTIMEO, return value is a float number"
@@ -185,6 +237,48 @@
                 len)
     (float (/ (fli:dereference timeout) 1000))))
 
+#+win32
+(defun get-socket-send-timeout (socket-fd)
+  "Get socket option: SNDTIMEO, return value is a float number"
+  (declare (type integer socket-fd))
+  (fli:with-dynamic-foreign-objects ((timeout :int)
+                                     (len :int))
+    (comm::getsockopt socket-fd
+                comm::*sockopt_sol_socket*
+                *sockopt_so_sndtimeo*
+                (fli:copy-pointer timeout
+                                  :type '(:pointer :void))
+                len)
+    (float (/ (fli:dereference timeout) 1000))))
+
+#+(or lispworks4 lispworks5.0)
+(defun set-socket-tcp-nodelay (socket-fd new-value)
+  "Set socket option: TCP_NODELAY, argument is a fixnum (0 or 1)"
+  (declare (type integer socket-fd)
+           (type (integer 0 1) new-value))
+  (fli:with-dynamic-foreign-objects ((zero-or-one :int))
+    (setf (fli:dereference zero-or-one) new-value)
+    (when (zerop (comm::setsockopt socket-fd
+                                   comm::*sockopt_sol_socket*
+                                   comm::*sockopt_tcp_nodelay*
+                                   (fli:copy-pointer zero-or-one
+                                                     :type '(:pointer #+win32 :char #-win32 :void))
+                                   (fli:size-of :int)))
+        new-value)))
+
+(defun get-socket-tcp-nodelay (socket-fd)
+  "Get socket option: TCP_NODELAY, return value is a fixnum (0 or 1)"
+  (declare (type integer socket-fd))
+  (fli:with-dynamic-foreign-objects ((zero-or-one :int)
+                                     (len :int))
+    (if (zerop (comm::getsockopt socket-fd
+                                 comm::*sockopt_sol_socket*
+                                 comm::*sockopt_tcp_nodelay*
+                                 (fli:copy-pointer zero-or-one
+                                                   :type '(:pointer #+win32 :char #-win32 :void))
+                                 len))
+        zero-or-one 0))) ; on error, return 0
+
 (defun initialize-dynamic-sockaddr (hostname service protocol &aux (original-hostname hostname))
   (declare (ignorable original-hostname))
   #+(or lispworks4 lispworks5 lispworks6.0)
@@ -198,7 +292,7 @@
             comm::*socket_af_inet*
             server-addr
             (fli:pointer-element-size server-addr)))
-  #-(or lispworks4 lispworks5 lispworks6.0)
+  #-(or lispworks4 lispworks5 lispworks6.0) ; version>=6.1
   (progn
     (when (stringp hostname)
       (setq hostname (comm:string-ip-address hostname))
@@ -221,8 +315,8 @@
                     server-addr
                     (if (eql family comm::*socket_af_inet*)
                         (fli:size-of '(:struct comm::sockaddr_in))
-                      (fli:size-of '(:struct comm::sockaddr_in6))))))
-      :bad-host)))
+			(fli:size-of '(:struct comm::sockaddr_in6))))))
+	:bad-host)))
 
 (defun open-udp-socket (&key local-address local-port read-timeout
                              (address-family comm::*socket_af_inet*))
@@ -290,9 +384,8 @@
           (error "cannot create socket"))))))
 
 (defun socket-connect (host port &key (protocol :stream) (element-type 'base-char)
-                       timeout deadline (nodelay t nodelay-specified)
+                       timeout deadline (nodelay t)
                        local-host local-port)
-
   ;; What's the meaning of this keyword?
   (when deadline
     (unimplemented 'deadline 'socket-connect))
@@ -301,11 +394,6 @@
   (when timeout
     (unsupported 'timeout 'socket-connect :minimum "LispWorks 4.4.5"))
 
-  #+(or lispworks4 lispworks5.0) ; < 5.1
-  (when (and nodelay-specified 
-             (not (eq nodelay :if-supported)))
-    (unsupported 'nodelay 'socket-connect :minimum "LispWorks 5.1"))
-
   #+lispworks4
   (when local-host
      (unsupported 'local-host 'socket-connect :minimum "LispWorks 5.0"))
@@ -317,7 +405,7 @@
     (:stream
      (let ((hostname (host-to-hostname host))
            (stream))
-       (setf stream
+       (setq stream
              (with-mapped-conditions ()
                (comm:open-tcp-stream hostname port
                                      :element-type element-type
@@ -331,6 +419,14 @@
                                      #-(or lispworks4 lispworks5.0) ; >= 5.1
                                      #-(or lispworks4 lispworks5.0)
                                      :nodelay nodelay)))
+
+       ;; Then handle `nodelay' separately for older versions <= 5.0
+       #+(or lispworks4 lispworks5.0)
+       (when (and stream nodelay)
+         (set-socket-tcp-nodelay
+           (comm:socket-stream-socket stream)
+           (bool->int nodelay))) ; ":if-supported" maps to 1 too.
+
        (if stream
            (make-stream-socket :socket (comm:socket-stream-socket stream)
                                :stream stream)
@@ -403,13 +499,9 @@
 ;; are correctly flushed and the socket closed.
 (defmethod socket-close ((usocket stream-usocket))
   "Close socket."
-  (when (wait-list usocket)
-     (remove-waiter (wait-list usocket) usocket))
   (close (socket-stream usocket)))
 
 (defmethod socket-close ((usocket usocket))
-  (when (wait-list usocket)
-     (remove-waiter (wait-list usocket) usocket))
   (with-mapped-conditions (usocket)
      (comm::close-socket (socket usocket))))
 
@@ -417,6 +509,45 @@
   "Additional socket-close method for datagram-usocket"
   (setf (%open-p socket) nil))
 
+(defconstant +shutdown-read+ 0)
+(defconstant +shutdown-write+ 1)
+(defconstant +shutdown-read-write+ 2)
+
+;;; int
+;;; shutdown(int socket, int what);
+(fli:define-foreign-function (%shutdown "shutdown" :source)
+    ((socket :int)
+     (what :int))
+  :result-type :int
+  #+win32 :module
+  #+win32 "ws2_32")
+
+(defmethod socket-shutdown ((usocket datagram-usocket) direction)
+  (unless (member direction '(:input :output :io))
+    (error 'invalid-argument-error))
+  (let ((what (case direction
+		(:input +shutdown-read+)
+		(:output +shutdown-write+)
+		(:io +shutdown-read-write+))))
+    (with-mapped-conditions (usocket)
+      #-(or lispworks4 lispworks5 lispworks6) ; lispworks 7.0+
+      (comm::shutdown (socket usocket) what)
+      #+(or lispworks4 lispworks5 lispworks6)
+      (= 0 (%shutdown (socket usocket) what)))))
+
+(defmethod socket-shutdown ((usocket stream-usocket) direction)
+  (unless (member direction '(:input :output :io))
+    (error 'invalid-argument-error))
+  (with-mapped-conditions (usocket)
+    #-(or lispworks4 lispworks5 lispworks6)
+    (comm:socket-stream-shutdown (socket usocket) direction)
+    #+(or lispworks4 lispworks5 lispworks6)
+    (let ((what (case direction
+		  (:input +shutdown-read+)
+		  (:output +shutdown-write+)
+		  (:io +shutdown-read-write+))))
+      (= 0 (%shutdown (comm:socket-stream-socket (socket usocket)) what)))))
+
 (defmethod initialize-instance :after ((socket datagram-usocket) &key)
   (setf (slot-value socket 'send-buffer)
         (make-array +max-datagram-packet-size+
@@ -432,7 +563,7 @@
 
 (defmethod socket-send ((usocket datagram-usocket) buffer size &key host port (offset 0)
                         &aux (socket-fd (socket usocket))
-                             (message (slot-value usocket 'send-buffer)))
+                             (message (slot-value usocket 'send-buffer))) ; TODO: multiple threads send together?
   "Send message to a socket, using sendto()/send()"
   (declare (type integer socket-fd)
            (type sequence buffer))
@@ -459,9 +590,8 @@
           (if (zerop errno)
               n
             (raise-usock-err errno socket-fd)))))))
-    
-(defun receive-message (socket-fd message &optional buffer (length (length buffer))
-			&key read-timeout (max-buffer-size +max-datagram-packet-size+))
+
+(defmethod socket-receive ((socket datagram-usocket) buffer length &key timeout (max-buffer-size +max-datagram-packet-size+))
   "Receive message from socket, read-timeout is a float number in seconds.
 
    This function will return 4 values:
@@ -469,12 +599,19 @@
    2. number of receive bytes
    3. remote address
    4. remote port"
-  (declare (type integer socket-fd)
-           (type sequence buffer))
-  (let (old-timeout)
+  (declare (values (simple-array (unsigned-byte 8) (*)) ; buffer
+                   (integer 0)                          ; size
+                   (unsigned-byte 32)                   ; host
+                   (unsigned-byte 16))                  ; port
+	   (type sequence buffer))
+  (let ((socket-fd (socket socket))
+        (message (slot-value socket 'recv-buffer)) ; TODO: how multiple threads do this in parallel?
+        (read-timeout timeout)
+        old-timeout)
+    (declare (type integer socket-fd))
     (fli:with-dynamic-foreign-objects ((client-addr (:struct comm::sockaddr_in))
                                        (len :int
-					    #-(or lispworks4 lispworks5.0) ; <= 5.0
+                                            #-(or lispworks4 lispworks5.0) ; <= 5.0
                                             :initial-element *length-of-sockaddr_in*))
       #+(or lispworks4 lispworks5.0) ; <= 5.0
       (setf (fli:dereference len) *length-of-sockaddr_in*)
@@ -490,6 +627,8 @@
           ;; restore old read timeout
           (when (and read-timeout (/= old-timeout read-timeout))
             (set-socket-receive-timeout socket-fd old-timeout))
+          ;; Frank James' patch: reset the %read-p for WAIT-FOR-INPUT
+          #+win32 (setf (%ready-p socket) nil)
           (if (plusp n)
               (values (if buffer
                           (replace buffer message
@@ -516,18 +655,6 @@
                   (values nil n 0 0)
                 (raise-usock-err errno socket-fd)))))))))
 
-(defmethod socket-receive ((socket datagram-usocket) buffer length &key timeout)
-  (declare (values (simple-array (unsigned-byte 8) (*)) ; buffer
-		   (integer 0)                          ; size
-		   (unsigned-byte 32)                   ; host
-		   (unsigned-byte 16)))                 ; port
-  (multiple-value-bind (buffer size host port)
-      (receive-message (socket socket)
-                       (slot-value socket 'recv-buffer)
-                       buffer length
-                       :read-timeout timeout)
-    (values buffer size host port)))
-
 (defmethod get-local-name ((usocket usocket))
   (multiple-value-bind
       (address port)
@@ -552,7 +679,21 @@
 (defmethod get-peer-port ((usocket stream-usocket))
   (nth-value 1 (get-peer-name usocket)))
 
+#-(or lispworks4 lispworks5 lispworks6.0) ; version>= 6.1
+(defun ipv6-address-p (hostname)
+  (when (stringp hostname)
+    (setq hostname (comm:string-ip-address hostname))
+    (unless hostname
+      (let ((resolved-hostname (comm:get-host-entry hostname :fields '(:address))))
+	(unless resolved-hostname
+	  (return-from ipv6-address-p nil))
+	(setq hostname resolved-hostname))))
+  (comm:ipv6-address-p hostname))
+
 (defun lw-hbo-to-vector-quad (hbo)
+  #+(or lispworks4 lispworks5 lispworks6.0)
+  (hbo-to-vector-quad hbo)
+  #-(or lispworks4 lispworks5 lispworks6.0) ; version>= 6.1
   (if (comm:ipv6-address-p hbo)
       (ipv6-host-to-vector (comm:ipv6-address-string hbo))
     (hbo-to-vector-quad hbo)))
