@@ -96,7 +96,7 @@
     (ecase protocol
       (:stream
        (let ((mcl-sock
-	      (openmcl-socket:make-socket :remote-host host
+	      (openmcl-socket:make-socket :remote-host (host-to-hostname host)
 					  :remote-port port
 					  :local-host local-host
 					  :local-port local-port
@@ -152,11 +152,13 @@
 ;; and their associated objects are represented
 ;; by the same object.
 (defmethod socket-close ((usocket usocket))
-  (when (wait-list usocket)
-     (remove-waiter (wait-list usocket) usocket))
   (with-mapped-conditions (usocket)
     (close (socket usocket))))
 
+(defmethod socket-shutdown ((usocket usocket) direction)
+  (with-mapped-conditions (usocket)
+    (openmcl-socket:shutdown (socket usocket) :direction direction)))
+
 #-ipv6
 (defmethod socket-send ((usocket datagram-usocket) buffer size &key host port (offset 0))
   (with-mapped-conditions (usocket)
