@@ -1,6 +1,4 @@
-;;;; -*- Mode: Lisp -*-
-;;;; $Id$
-;;;; $URL$
+;;;; -*- Mode: Common-Lisp -*-
 
 ;;;; See LICENSE for licensing information.
 
@@ -32,7 +30,8 @@
   #-:wsock
   (ffi:clines
    "#include <errno.h>"
-   "#include <sys/socket.h>")
+   "#include <sys/socket.h>"
+   "#include <unistd.h>")
   #+:wsock
   (ffi:clines
    "#ifndef FD_SETSIZE"
@@ -54,6 +53,11 @@
    "#define CONS(x, y) make_cons((x), (y))"
    "#define MAKE_INTEGER(x) make_integer((x))")
 |#
+
+  (defun cerrno ()
+    (ffi:c-inline () () :int
+     "errno" :one-liner t))
+
   (defun fd-setsize ()
     (ffi:c-inline () () :fixnum
      "FD_SETSIZE" :one-liner t))
@@ -78,7 +82,8 @@
     (ffi:c-inline (fdset fd) (:pointer-void :fixnum) :bool
      "FD_ISSET(#1,(fd_set*)#0)" :one-liner t))
 
-  (declaim (inline fd-setsize
+  (declaim (inline cerrno
+                   fd-setsize
                    fdset-alloc
                    fdset-zero
                    fdset-set
@@ -88,7 +93,7 @@
   (defun get-host-name ()
     (ffi:c-inline
      () () :object
-     "{ char *buf = ecl_alloc_atomic(257);
+     "{ char *buf = (char *) ecl_alloc_atomic(257);
 
         if (gethostname(buf,256) == 0)
           @(return) = make_simple_base_string(buf);
@@ -130,7 +135,7 @@
            (values nil nil))
           ((< count 0)
            ;; check for EINTR and EAGAIN; these should not err
-           (values nil (ffi:c-inline () () :int "errno" :one-liner t)))
+           (values nil (cerrno)))
           (t
            (dolist (sock sockets)
              (when (fdset-fd-isset rfds (sb-bsd-sockets:socket-file-descriptor
@@ -171,21 +176,23 @@
     (sb-bsd-sockets:socket-error . ,#'map-socket-error)
 
     ;; Nameservice errors: mapped to unknown-error
-    #-ecl #-ecl #-ecl
+    #-ecl
     (sb-bsd-sockets:no-recovery-error . ns-no-recovery-error)
+    #-ecl
     (sb-bsd-sockets:try-again-error . ns-try-again-condition)
+    #-ecl
     (sb-bsd-sockets:host-not-found-error . ns-host-not-found-error)))
 
 (defun handle-condition (condition &optional (socket nil))
   "Dispatch correct usocket condition."
   (typecase condition
-    (serious-condition (let* ((usock-error (cdr (assoc (type-of condition)
-                                           +sbcl-error-map+)))
-                  (usock-error (if (functionp usock-error)
-                                   (funcall usock-error condition)
-                                 usock-error)))
-             (when usock-error
-                 (error usock-error :socket socket))))
+    (serious-condition
+     (let* ((usock-error (cdr (assoc (type-of condition) +sbcl-error-map+)))
+	    (usock-error (if (functionp usock-error)
+			     (funcall usock-error condition)
+			   usock-error)))
+       (when usock-error
+	 (error usock-error :socket socket))))
     (condition (let* ((usock-cond (cdr (assoc (type-of condition)
                                               +sbcl-condition-map+)))
                       (usock-cond (if (functionp usock-cond)
@@ -240,6 +247,16 @@ happen. Use with care."
 	  ,timeout
 	  (return-from ,block ,timeout-form)))))
 
+(defun get-hosts-by-name (name)
+  (with-mapped-conditions ()
+    (multiple-value-bind (host4 host6)
+        (sb-bsd-sockets:get-host-by-name name)
+      (let ((addr4 (when host4
+                     (sb-bsd-sockets::host-ent-addresses host4)))
+            (addr6 (when host6
+                     (sb-bsd-sockets::host-ent-addresses host6))))
+        (append addr4 addr6)))))
+
 (defun socket-connect (host port &key (protocol :stream) (element-type 'character)
                        timeout deadline (nodelay t nodelay-specified)
                        local-host local-port
@@ -260,12 +277,23 @@ happen. Use with care."
   (when (eq nodelay :if-supported)
     (setf nodelay t))
 
-  (let ((socket (make-instance 'sb-bsd-sockets:inet-socket
-                               :type protocol
-                               :protocol (case protocol
-                                           (:stream :tcp)
-                                           (:datagram :udp))))
-        usocket ok)
+  (let* ((remote (when host
+                   (car (get-hosts-by-name (host-to-hostname host)))))
+         (local (when local-host
+                  (car (get-hosts-by-name (host-to-hostname local-host)))))
+         (ipv6 (or (and remote (= 16 (length remote)))
+                   (and local (= 16 (length local)))))
+         (socket (make-instance #+sbcl (if ipv6
+                                           'sb-bsd-sockets::inet6-socket
+                                           'sb-bsd-sockets:inet-socket)
+                                #+ecl 'sb-bsd-sockets:inet-socket
+                                :type protocol
+                                :protocol (case protocol
+                                            (:stream :tcp)
+                                            (:datagram :udp))))
+         usocket
+         ok)
+
     (unwind-protect
          (progn
            (ecase protocol
@@ -280,18 +308,20 @@ happen. Use with care."
                 (setf (sb-bsd-sockets::sockopt-tcp-nodelay socket) nodelay))
               (when (or local-host local-port)
                 (sb-bsd-sockets:socket-bind socket
-					    (host-to-vector-quad
-					     (or local-host *wildcard-host*))
+                                            (if ipv6
+                                                (or local (ipv6-host-to-vector "::0"))
+                                                (or local (host-to-vector-quad *wildcard-host*)))
                                             (or local-port *auto-port*)))
+
               (with-mapped-conditions (usocket)
 		#+(and sbcl (not win32))
 		(labels ((connect ()
-			   (sb-bsd-sockets:socket-connect socket (host-to-vector-quad host) port)))
+                           (sb-bsd-sockets:socket-connect socket remote port)))
 		  (if timeout
 		      (%with-timeout (timeout (error 'sb-ext:timeout)) (connect))
 		      (connect)))
 		#+(or ecl (and sbcl win32))
-		(sb-bsd-sockets:socket-connect socket (host-to-vector-quad host) port)
+		(sb-bsd-sockets:socket-connect socket remote port)
                 ;; Now that we're connected make the stream.
                 (setf (socket-stream usocket)
                       (sb-bsd-sockets:socket-make-stream socket
@@ -316,13 +346,14 @@ happen. Use with care."
              (:datagram
               (when (or local-host local-port)
                 (sb-bsd-sockets:socket-bind socket
-                                            (host-to-vector-quad
-                                             (or local-host *wildcard-host*))
+                                            (if ipv6
+                                                (or local (ipv6-host-to-vector "::0"))
+                                                (or local (host-to-vector-quad *wildcard-host*)))
                                             (or local-port *auto-port*)))
               (setf usocket (make-datagram-socket socket))
               (when (and host port)
                 (with-mapped-conditions (usocket)
-                  (sb-bsd-sockets:socket-connect socket (host-to-vector-quad host) port)
+                  (sb-bsd-sockets:socket-connect socket remote port)
                   (setf (connected-p usocket) t)))))
            (setf ok t))
       ;; Clean up in case of an error.
@@ -335,15 +366,22 @@ happen. Use with care."
                            (reuse-address nil reuse-address-supplied-p)
                            (backlog 5)
                            (element-type 'character))
-  (let* ((reuseaddress (if reuse-address-supplied-p reuse-address reuseaddress))
-         (ip #+sbcl
-	     (if (and host
-		      (not (eq host *wildcard-host*)))
-		 (host-to-vector-quad host)
-		 (hbo-to-vector-quad sb-bsd-sockets-internal::inaddr-any))
-	     #-sbcl (host-to-vector-quad host))
-         (sock (make-instance 'sb-bsd-sockets:inet-socket
-                              :type :stream :protocol :tcp)))
+  (let* (#+sbcl
+	 (local (when host
+                  (car (get-hosts-by-name (host-to-hostname host)))))
+	 #+sbcl
+         (ipv6 (and local (= 16 (length local))))
+         (reuseaddress (if reuse-address-supplied-p reuse-address reuseaddress))
+         (ip #+sbcl (if (and local (not (eq host *wildcard-host*)))
+                        local
+                        (hbo-to-vector-quad sb-bsd-sockets-internal::inaddr-any))
+             #+ecl (host-to-vector-quad host))
+         (sock (make-instance #+sbcl (if ipv6
+                                         'sb-bsd-sockets::inet6-socket
+                                         'sb-bsd-sockets:inet-socket)
+                              #+ecl 'sb-bsd-sockets:inet-socket
+                              :type :stream
+                              :protocol :tcp)))
     (handler-case
         (with-mapped-conditions ()
           (setf (sb-bsd-sockets:sockopt-reuse-address sock) reuseaddress)
@@ -385,31 +423,45 @@ happen. Use with care."
 ;; different objects. Be sure to close the stream (which
 ;; closes the socket too) when closing a stream-socket.
 (defmethod socket-close ((usocket usocket))
-  (when (wait-list usocket)
-     (remove-waiter (wait-list usocket) usocket))
   (with-mapped-conditions (usocket)
     (sb-bsd-sockets:socket-close (socket usocket))))
 
 (defmethod socket-close ((usocket stream-usocket))
-  (when (wait-list usocket)
-     (remove-waiter (wait-list usocket) usocket))
   (with-mapped-conditions (usocket)
     (close (socket-stream usocket))))
 
-(defmethod socket-send ((usocket datagram-usocket) buffer size &key host port (offset 0))
+#+sbcl
+(defmethod socket-shutdown ((usocket stream-usocket) direction)
   (with-mapped-conditions (usocket)
-    (let* ((s (socket usocket))
-	   (dest (if (and host port) (list (host-to-vector-quad host) port) nil))
-	   (real-buffer (if (zerop offset)
-			    buffer
-			    (subseq buffer offset (+ offset size)))))
-      (sb-bsd-sockets:socket-send s real-buffer size :address dest))))
+    (sb-bsd-sockets::socket-shutdown (socket usocket) :direction direction)))
+
+#+ecl
+(defmethod socket-shutdown ((usocket stream-usocket) direction)
+  (let ((sock-fd (sb-bsd-sockets:socket-file-descriptor (socket usocket)))
+        (direction-flag (ecase direction
+                          (:input 0)
+                          (:output 1))))
+    (unless (zerop (ffi:c-inline (sock-fd direction-flag) (:int :int) :int
+                               "shutdown(#0, #1)" :one-liner t))
+      (error (map-errno-error (cerrno))))))
+
+(defmethod socket-send ((usocket datagram-usocket) buffer size &key host port (offset 0))
+  (let ((remote (when host
+                  (car (get-hosts-by-name (host-to-hostname host))))))
+    (with-mapped-conditions (usocket)
+      (let* ((s (socket usocket))
+             (dest (if (and host port) (list remote port) nil))
+             (real-buffer (if (zerop offset)
+                              buffer
+                              (subseq buffer offset (+ offset size)))))
+        (sb-bsd-sockets:socket-send s real-buffer size :address dest)))))
 
 (defmethod socket-receive ((socket datagram-usocket) buffer length
 			   &key (element-type '(unsigned-byte 8)))
+  #+sbcl
   (declare (values (simple-array (unsigned-byte 8) (*)) ; buffer
 		   (integer 0)                          ; size
-		   (unsigned-byte 32)                   ; host
+		   (simple-array (unsigned-byte 8) (*)) ; host
 		   (unsigned-byte 16)))                 ; port
   (with-mapped-conditions (socket)
     (let ((s (socket socket)))
@@ -438,11 +490,6 @@ happen. Use with care."
     (sb-bsd-sockets::host-ent-name
         (sb-bsd-sockets:get-host-by-address address))))
 
-(defun get-hosts-by-name (name)
-  (with-mapped-conditions ()
-     (sb-bsd-sockets::host-ent-addresses
-         (sb-bsd-sockets:get-host-by-name name))))
-
 #+(and sbcl (not win32))
 (progn
   (defun %setup-wait-list (wait-list)
@@ -466,18 +513,22 @@ happen. Use with care."
          (multiple-value-bind
              (secs musecs)
              (split-timeout (or timeout 1))
-           (multiple-value-bind
-               (count err)
-               (sb-unix:unix-fast-select
-                (1+ (reduce #'max (wait-list-%wait sockets)
-                            :key #'sb-bsd-sockets:socket-file-descriptor))
-                (sb-alien:addr rfds) nil nil
-                (when timeout secs) (when timeout musecs))
-	     (if (null count)
-		 (unless (= err sb-unix:EINTR)
-		   (error (map-errno-error err)))
-		 (when (< 0 count)
-		   ;; process the result...
+           (let* ((wait-list (wait-list-%wait sockets))
+                  count err)
+             (if (null wait-list)
+                 (setq count 0) ;; no need to call
+               (multiple-value-setq (count err)
+                 (sb-unix:unix-fast-select
+                  ;; "invalid number of arguments: 0" if wait-list is null.
+                  (1+ (reduce #'max wait-list
+                              :key #'sb-bsd-sockets:socket-file-descriptor))
+                  (sb-alien:addr rfds) nil nil
+                  (when timeout secs) (when timeout musecs))))
+             (if (null count) ; something wrong in #'sb-unix:unix-fast-select
+                 (unless (= err sb-unix:EINTR)
+                   (error (map-errno-error err)))
+                 (when (< 0 count) ; do nothing if count = 0
+                   ;; process the result...
                    (dolist (x (wait-list-waiters sockets))
                      (when (sb-unix:fd-isset
                             (sb-bsd-sockets:socket-file-descriptor
@@ -492,6 +543,7 @@ happen. Use with care."
 #+win32 ; shared by ECL and SBCL
 (eval-when (:compile-toplevel :load-toplevel :execute)
   (defconstant +wsa-wait-failed+ #xffffffff)
+  (defconstant +wsa-infinite+ #xffffffff)
   (defconstant +wsa-wait-event-0+ 0)
   (defconstant +wsa-wait-timeout+ 258))
 
@@ -541,7 +593,11 @@ happen. Use with care."
   (defun wait-for-input-internal (wait-list &key timeout)
     (when (waiting-required (wait-list-waiters wait-list))
       (let ((rv (wsa-wait-for-multiple-events 1 (wait-list-%wait wait-list)
-                                              nil (truncate (* 1000 timeout)) nil)))
+                                              nil
+                                              (if timeout
+                                                  (truncate (* 1000 timeout))
+                                                  +wsa-infinite+)
+                                              nil)))
         (ecase rv
           ((#.+wsa-wait-event-0+)
            (update-ready-and-state-slots (wait-list-waiters wait-list)))
@@ -699,12 +755,11 @@ happen. Use with care."
 (progn
   (defun wait-for-input-internal (wl &key timeout)
     (with-mapped-conditions ()
-      (multiple-value-bind
-            (secs usecs)
+      (multiple-value-bind (secs usecs)
           (split-timeout (or timeout 1))
-        (multiple-value-bind
-              (result-fds err)
+        (multiple-value-bind (result-fds err)
             (read-select wl (when timeout secs) usecs)
+	  (declare (ignore result-fds))
           (unless (null err)
             (error (map-errno-error err)))))))
 
@@ -801,7 +856,10 @@ happen. Use with care."
 
   (defun wait-for-input-internal (wait-list &key timeout)
     (when (waiting-required (wait-list-waiters wait-list))
-      (let ((rv (ffi:c-inline ((wait-list-%wait wait-list) (truncate (* 1000 timeout)))
+      (let ((rv (ffi:c-inline ((wait-list-%wait wait-list)
+                               (if timeout
+                                   (truncate (* 1000 timeout))
+                                   +wsa-infinite+))
                               (:fixnum :fixnum) :fixnum
                  "DWORD result;
                   WSAEVENT events[1];
