@@ -1,7 +1,5 @@
-;;;; $Id: wait-for-input.lisp 656 2011-05-11 07:05:39Z ctian $
-;;;; $URL: svn://common-lisp.net/project/usocket/svn/usocket/tags/0.5.5/test/wait-for-input.lisp $
-
 ;;;; See LICENSE for licensing information.
+
 (in-package :usocket-test)
 
 (eval-when (:compile-toplevel :load-toplevel :execute)
@@ -36,9 +34,9 @@
                     "GET / HTTP/1.0~2%")
             (force-output (usocket:socket-stream sock))
             (usocket:wait-for-input sock :timeout *wait-for-input-timeout*)
-            (subseq (read-line (usocket:socket-stream sock)) 0 15))
+            (subseq (read-line (usocket:socket-stream sock)) 0 4))
         (usocket:socket-close sock))))
-  "HTTP/1.1 200 OK")
+  "HTTP")
 
 ;;; Advanced W-F-I tests by Elliott Slaughter <elliottslaughter@gmail.com>
 
