@@ -1,16 +1,18 @@
-;;;; $Id$
-;;;; $URL$
-
 ;;;; See LICENSE for licensing information.
 
 (in-package :usocket)
 
 (defparameter *wildcard-host* #(0 0 0 0)
-  "Hostname to pass when all interfaces in the current system are to be bound.")
+  "Hostname to pass when all interfaces in the current system are to
+  be bound.  If this variable is passed to socket-listen, IPv6 capable
+  systems will also listen for IPv6 connections.")
 
 (defparameter *auto-port* 0
   "Port number to pass when an auto-assigned port number is wanted.")
 
+(defparameter *version* #.(asdf:component-version (asdf:find-system :usocket))
+  "usocket version string")
+
 (defconstant +max-datagram-packet-size+ 65507
   "The theoretical maximum amount of data in a UDP datagram.
 
@@ -67,6 +69,15 @@ Note: Accessed, but not used for 'stream-usocket'.
 
 Sockets should be closed using the `socket-close' method."))
 
+(defgeneric socket-state (socket)
+  (:documentation "NIL          - not ready
+:READ        - ready to read
+:READ-WRITE  - ready to read and write
+:WRITE       - ready to write"))
+
+(defmethod socket-state ((socket usocket))
+  (state socket))
+
 (defclass stream-usocket (usocket)
    ((stream
      :initarg :stream
@@ -177,6 +188,19 @@ explicitly specified, or the element-type passed to `socket-listen' otherwise.")
 (defgeneric socket-close (usocket)
   (:documentation "Close a previously opened `usocket'."))
 
+(defmethod socket-close :before ((usocket usocket))
+  (when (wait-list usocket)
+    (remove-waiter (wait-list usocket) usocket)))
+
+;; also see http://stackoverflow.com/questions/4160347/close-vs-shutdown-socket
+(defgeneric socket-shutdown (usocket direction)
+  (:documentation "Shutdown communication on the socket in DIRECTION.
+
+After a shutdown no input and/or output of the indicated DIRECTION
+can be performed on the `usocket'.
+
+DIRECTION should be either :INPUT or :OUTPUT or :IO"))
+
 (defgeneric socket-send (usocket buffer length &key host port)
   (:documentation "Send packets through a previously opend `usocket'."))
 
@@ -268,7 +292,7 @@ The `body' is an implied progn form."
   (let ((wl (%make-wait-list)))
     (setf (wait-list-map wl) (make-hash-table))
     (%setup-wait-list wl)
-    (dolist (x waiters wl)
+    (dolist (x waiters wl) ; wl is returned
       (add-waiter wl x))))
 
 (defun add-waiter (wait-list input)
@@ -290,7 +314,8 @@ The `body' is an implied progn form."
   (setf (wait-list-waiters wait-list) nil)
   (clrhash (wait-list-map wait-list)))
 
-(defun wait-for-input (socket-or-sockets &key timeout ready-only)
+(defun wait-for-input (socket-or-sockets &key timeout ready-only
+                                         &aux (single-socket-p (atom socket-or-sockets)))
   "Waits for one or more streams to become ready for reading from
 the socket.  When `timeout' (a non-negative real number) is
 specified, wait `timeout' seconds, or wait indefinitely when
@@ -311,14 +336,30 @@ to do the associated garbage collection.
 
 Without the READY-ONLY arg, you need to check the socket STATE slot for
 the values documented in usocket.lisp in the usocket class."
+
+  ;; for NULL sockets, return NIL with respect of TIMEOUT.
+  (when (null socket-or-sockets)
+    (when timeout
+      (sleep timeout))
+    (return-from wait-for-input nil))
+
+  ;; create a new wait-list if it's not created by the caller.
   (unless (wait-list-p socket-or-sockets)
-    (let ((wl (make-wait-list (if (listp socket-or-sockets)
-                                  socket-or-sockets (list socket-or-sockets)))))
+    ;; OPTIMIZATION: in case socket-or-sockets is an atom, create the wait-list
+    ;; only once and store it into the usocket itself.   
+    (let ((wl (if (and single-socket-p (wait-list socket-or-sockets))
+                  (wait-list socket-or-sockets) ; reuse the wait-list
+                (make-wait-list (if (listp socket-or-sockets)
+                                    socket-or-sockets (list socket-or-sockets))))))
       (multiple-value-bind
-            (socks to)
+            (sockets to-result)
           (wait-for-input wl :timeout timeout :ready-only ready-only)
+        ;; in case of single socket, keep the wait-list
+        (unless single-socket-p
+          (remove-all-waiters wl))
         (return-from wait-for-input
-          (values (if ready-only socks socket-or-sockets) to)))))
+          (values (if ready-only sockets socket-or-sockets) to-result)))))
+
   (let* ((start (get-internal-real-time))
          (sockets-ready 0))
     (dolist (x (wait-list-waiters socket-or-sockets))
@@ -333,7 +374,7 @@ the values documented in usocket.lisp in the usocket class."
     ;; the internal routine is responsibe for
     ;; making sure the wait doesn't block on socket-streams of
     ;; which theready- socket isn't ready, but there's space left in the
-    ;; buffer
+    ;; buffer.  socket-or-sockets is not destructed.
     (wait-for-input-internal socket-or-sockets
                              :timeout (if (zerop sockets-ready) timeout 0))
     (let ((to-result (when timeout
@@ -341,9 +382,16 @@ the values documented in usocket.lisp in the usocket class."
                                          internal-time-units-per-second)))
                          (when (< elapsed timeout)
                            (- timeout elapsed))))))
-      (values (if ready-only
-                  (remove-if #'null (wait-list-waiters socket-or-sockets) :key #'state)
-                  socket-or-sockets)
+      ;; two return values:
+      ;; 1) the original wait-list, or available sockets (ready-only)
+      ;; 2) remaining timeout
+      (values (cond (ready-only
+                     (cond (single-socket-p
+                            (if (null (state (car (wait-list-waiters socket-or-sockets))))
+                                nil ; nothing left if the only socket is not waiting
+                              (wait-list-waiters socket-or-sockets)))
+                           (t (remove-if #'null (wait-list-waiters socket-or-sockets) :key #'state))))
+                    (t socket-or-sockets))
               to-result))))
 
 ;;
@@ -381,7 +429,7 @@ the values documented in usocket.lisp in the usocket class."
   `(octet-buffer-to-integer ,buffer 4 :start ,start))
 
 ;;
-;; IP(v4) utility functions
+;; IPv4 utility functions
 ;;
 
 (defun list-of-strings-to-integers (list)
@@ -399,7 +447,7 @@ parse-integer) on each of the string elements."
                (eql char #\.)))
          string))
 
-(defun hbo-to-dotted-quad (integer)
+(defun hbo-to-dotted-quad (integer) ; exported
   "Host-byte-order integer to dotted-quad string conversion utility."
   (let ((first (ldb (byte 8 24) integer))
         (second (ldb (byte 8 16) integer))
@@ -407,7 +455,7 @@ parse-integer) on each of the string elements."
         (fourth (ldb (byte 8 0) integer)))
     (format nil "~A.~A.~A.~A" first second third fourth)))
 
-(defun hbo-to-vector-quad (integer)
+(defun hbo-to-vector-quad (integer) ; exported
   "Host-byte-order integer to dotted-quad string conversion utility."
   (let ((first (ldb (byte 8 24) integer))
         (second (ldb (byte 8 16) integer))
@@ -415,18 +463,19 @@ parse-integer) on each of the string elements."
         (fourth (ldb (byte 8 0) integer)))
     (vector first second third fourth)))
 
-(defun vector-quad-to-dotted-quad (vector)
+(defun vector-quad-to-dotted-quad (vector) ; exported
   (format nil "~A.~A.~A.~A"
           (aref vector 0)
           (aref vector 1)
           (aref vector 2)
           (aref vector 3)))
 
-(defun dotted-quad-to-vector-quad (string)
+(defun dotted-quad-to-vector-quad (string) ; exported
   (let ((list (list-of-strings-to-integers (split-sequence #\. string))))
     (vector (first list) (second list) (third list) (fourth list))))
 
-(defgeneric host-byte-order (address))
+(defgeneric host-byte-order (address)) ; exported
+
 (defmethod host-byte-order ((string string))
   "Convert a string, such as 192.168.1.1, to host-byte-order,
 such as 3232235777."
@@ -434,38 +483,124 @@ such as 3232235777."
     (+ (* (first list) 256 256 256) (* (second list) 256 256)
        (* (third list) 256) (fourth list))))
 
-(defmethod host-byte-order ((vector vector))
+(defmethod host-byte-order ((vector vector)) ; IPv4 only
   "Convert a vector, such as #(192 168 1 1), to host-byte-order, such as
 3232235777."
   (+ (* (aref vector 0) 256 256 256) (* (aref vector 1) 256 256)
      (* (aref vector 2) 256) (aref vector 3)))
 
 (defmethod host-byte-order ((int integer))
-  int)
+  int) ; this assume input integer is already host-byte-order
 
-(defun host-to-hostname (host)
-  "Translate a string or vector quad to a stringified hostname."
+;;
+;; IPv6 utility functions
+;;
+
+(defun vector-to-ipv6-host (vector) ; exported
+  (with-output-to-string (*standard-output*)
+    (loop with zeros-collapsed-p
+          with collapsing-zeros-p
+          for i below 16 by 2
+          for word = (+ (ash (aref vector i) 8)
+                        (aref vector (1+ i)))
+          do (cond
+               ((and (zerop word)
+                     (not collapsing-zeros-p)
+                     (not zeros-collapsed-p))
+                (setf collapsing-zeros-p t))
+               ((or (not (zerop word))
+                    zeros-collapsed-p)
+                (when collapsing-zeros-p
+                  (write-string ":")
+                  (setf collapsing-zeros-p nil
+                        zeros-collapsed-p t))
+                (format t "~:[~;:~]~X" (plusp i) word)))
+          finally (when collapsing-zeros-p
+                    (write-string "::")))))
+
+(defun split-ipv6-address (string)
+  (let ((pos 0)
+        word
+        double-colon-seen-p
+        words-before-double-colon
+        words-after-double-colon)
+    (loop
+      (multiple-value-setq (word pos) (parse-integer string :radix 16 :junk-allowed t :start pos))
+      (labels ((at-end-p ()
+                 (= pos (length string)))
+               (looking-at-colon-p ()
+                 (char= (char string pos) #\:))
+               (ensure-colon ()
+                 (unless (looking-at-colon-p)
+                   (error "unsyntactic IPv6 address string ~S, expected a colon at position ~D"
+                          string pos))
+                 (incf pos)))
+        (cond
+          ((null word)
+           (when double-colon-seen-p
+             (error "unsyntactic IPv6 address string ~S, can only have one double-colon filler mark"
+                    string))
+           (setf double-colon-seen-p t))
+          (double-colon-seen-p
+           (push word words-after-double-colon))
+          (t
+           (push word words-before-double-colon)))
+        (if (at-end-p)
+            (return (list (nreverse words-before-double-colon) (nreverse words-after-double-colon)))
+            (ensure-colon))))))
+
+(defun ipv6-host-to-vector (string) ; exported
+  (assert (> (length string) 2) ()
+          "Unsyntactic IPv6 address literal ~S, expected at least three characters" string)
+  (destructuring-bind (words-before-double-colon words-after-double-colon)
+      (split-ipv6-address (concatenate 'string
+                                       (when (eql (char string 0) #\:)
+                                         "0")
+                                       string
+                                       (when (eql (char string (1- (length string))) #\:)
+                                         "0")))
+    (let ((number-of-words-specified (+ (length words-before-double-colon) (length words-after-double-colon))))
+      (assert (<= number-of-words-specified 8) ()
+              "Unsyntactic IPv6 address literal ~S, too many colon separated address components" string)
+      (assert (or (= number-of-words-specified 8) words-after-double-colon) ()
+              "Unsyntactic IPv6 address literal ~S, too few address components and no double-colon filler found" string)
+      (loop with vector = (make-array 16 :element-type '(unsigned-byte 8))
+            for i below 16 by 2
+            for word in (append words-before-double-colon
+                                (make-list (- 8 number-of-words-specified) :initial-element 0)
+                                words-after-double-colon)
+            do (setf (aref vector i) (ldb (byte 8 8) word)
+                     (aref vector (1+ i)) (ldb (byte 8 0) word))
+            finally (return vector)))))
+
+;; exported since 0.8.0
+(defun host-to-hostname (host) ; host -> string
+  "Translate a string, vector quad or 16 byte IPv6 address to a
+stringified hostname."
   (etypecase host
-    (string host)
-    ((or (vector t 4)
+    (string host)      ; IPv4 or IPv6
+    ((or (vector t 4)  ; IPv4
          (array (unsigned-byte 8) (4)))
      (vector-quad-to-dotted-quad host))
-    (integer (hbo-to-dotted-quad host))
-    (null "0.0.0.0")))
+    ((or (vector t 16) ; IPv6
+         (array (unsigned-byte 8) (16)))
+     (vector-to-ipv6-host host))
+    (integer (hbo-to-dotted-quad host)) ; integer input is IPv4 only
+    (null "0.0.0.0")))                  ; null is IPv4
 
-(defun ip= (ip1 ip2)
+(defun ip= (ip1 ip2) ; exported
   (etypecase ip1
-    (string (string= ip1 (host-to-hostname ip2)))
-    ((or (vector t 4)
-         (array (unsigned-byte 8) (4)))
-     (or (eq ip1 ip2)
-         (and (= (aref ip1 0) (aref ip2 0))
-              (= (aref ip1 1) (aref ip2 1))
-              (= (aref ip1 2) (aref ip2 2))
-              (= (aref ip1 3) (aref ip2 3)))))
-    (integer (= ip1 (host-byte-order ip2)))))
+    (string (string= ip1                  ; IPv4 or IPv6
+                     (host-to-hostname ip2)))
+    ((or (vector t 4)                     ; IPv4
+         (array (unsigned-byte 8) (4))    ; IPv4
+         (vector t 16)                    ; IPv6
+         (array (unsigned-byte 8) (16)))  ; IPv6
+     (equalp ip1 ip2))
+    (integer (= ip1                       ; IPv4 only
+                (host-byte-order ip2))))) ; convert ip2 to integer (hbo)
 
-(defun ip/= (ip1 ip2)
+(defun ip/= (ip1 ip2) ; exported
   (not (ip= ip1 ip2)))
 
 ;;
@@ -473,22 +608,29 @@ such as 3232235777."
 ;;
 
 (defun get-host-by-name (name)
-  (let ((hosts (get-hosts-by-name name)))
-    (car hosts)))
+  "0.7.1+: if there're IPv4 addresses, return the first IPv4 address."
+  (let* ((hosts (get-hosts-by-name name))
+         (pos (position-if #'(lambda (ip) (= 4 (length ip))) hosts)))
+    (if pos (elt hosts pos)
+      (car hosts))))
 
 (defun get-random-host-by-name (name)
-  (let ((hosts (get-hosts-by-name name)))
-    (when hosts
-      (elt hosts (random (length hosts))))))
+  "0.7.1+: if there're IPv4 addresses, only return a random IPv4 address."
+  (let* ((hosts (get-hosts-by-name name))
+         (ipv4-hosts (remove-if-not #'(lambda (ip) (= 4 (length ip))) hosts)))
+    (cond (ipv4-hosts
+           (elt ipv4-hosts (random (length ipv4-hosts))))
+          (hosts
+           (elt hosts (random (length hosts)))))))
 
-(defun host-to-vector-quad (host)
+(defun host-to-vector-quad (host) ; internal
   "Translate a host specification (vector quad, dotted quad or domain name)
 to a vector quad."
   (etypecase host
     (string (let* ((ip (when (ip-address-string-p host)
                          (dotted-quad-to-vector-quad host))))
               (if (and ip (= 4 (length ip)))
-                  ;; valid IP dotted quad?
+                  ;; valid IP dotted quad? not sure
                   ip
                 (get-random-host-by-name host))))
     ((or (vector t 4)
@@ -496,7 +638,7 @@ to a vector quad."
      host)
     (integer (hbo-to-vector-quad host))))
 
-(defun host-to-hbo (host)
+(defun host-to-hbo (host) ; internal
   (etypecase host
     (string (let ((ip (when (ip-address-string-p host)
                         (dotted-quad-to-vector-quad host))))
@@ -538,15 +680,15 @@ an IP address represented in vector notation, such as #(192 168 1 1).
 stream associated with the socket.  The default is 'character.
 
 `nodelay' Allows to disable/enable Nagle's algorithm (http://en.wikipedia.org/wiki/Nagle%27s_algorithm).
-If this parameter is omitted, the behaviour is inherited from the 
-CL implementation (in most cases, Nagle's algorithm is 
+If this parameter is omitted, the behaviour is inherited from the
+CL implementation (in most cases, Nagle's algorithm is
 enabled by default, but for example in ACL it is disabled).
-If the parmeter is specified, one of these three values is possible: 
+If the parameter is specified, one of these three values is possible:
   T - Disable Nagle's algorithm; signals an UNSUPPORTED
-      condition if the implementation does not support explicit 
+      condition if the implementation does not support explicit
       manipulation with that option.
   NIL - Leave Nagle's algorithm enabled on the socket;
-      signals an UNSUPPORTED condition if the implementation does 
+      signals an UNSUPPORTED condition if the implementation does
       not support explicit manipulation with that option.
   :IF-SUPPORTED - Disables Nagle's algorithm if the implementation
       allows this, otherwises just ignore this option.
@@ -559,11 +701,11 @@ Returns a usocket object.")
 ;;###FIXME: extend with default-element-type
 (setf (documentation 'socket-listen 'function)
       "Bind to interface `host' on `port'. `host' should be the
-representation of an ready-interface address.  The implementation is not
-required to do an address lookup, making no guarantees that hostnames
-will be correctly resolved.  If `*wildcard-host*' is passed for `host',
-the socket will be bound to all available interfaces for the IPv4
-protocol in the system.  `port' can be selected by the IP stack by
+representation of an ready-interface address.  The implementation is
+not required to do an address lookup, making no guarantees that
+hostnames will be correctly resolved.  If `*wildcard-host*' or NIL is
+passed for `host', the socket will be bound to all available
+interfaces for the system.  `port' can be selected by the IP stack by
 passing `*auto-port*'.
 
 Returns an object of type `stream-server-usocket'.
@@ -574,3 +716,10 @@ streams to be created by `socket-accept'.  `reuseaddress' is supported for
 backward compatibility (but deprecated); when both `reuseaddress' and
 `reuse-address' have been specified, the latter takes precedence.
 ")
+
+;;; Small utility functions mapping true/false to 1/0, moved here from option.lisp
+
+(proclaim '(inline bool->int int->bool))
+
+(defun bool->int (bool) (if bool 1 0))
+(defun int->bool (int) (= 1 int))
